' Checker-Kit DIY draughts program, (C) PC SOLUTIONS, All rights reserved.
'  (Also called BASIC CHECKERS)
'   PC Solutions, PO Box 954, Bournemouth BH7 6YJ, England.

'   This is a source code of a fully functional, if not very strong
' program to play English draughts. You can run it under the QBASIC
' interpreter supplied with MSDOS5 - just type: QBASIC CHKIT11.BAS
' from the DOS prompt. For serious programming you should buy the
' MICROSOFT QUICKBASIC compiler, which will run the program about 15 times
' faster - when compiled to an EXE file.
'   I do NOT have space here to cover elementary BASIC programming
' methods - I suggest you buy a book on QUICKBASIC for information -
' however even non-programmers can have fun by altering the opening
' book statements at the end of this program.
'   Please remember that this code is Copyright A.Millett, PC Solutions,
' and may NOT be distributed - it is only for personal use of purchasers
' of our special Deluxe Draughts games-packs. However I realise that you
' may want to show off results to friends, so I give permission for you
' to distribute COMPILED .EXE files derived from this BASIC program
' so long as it is NOT sold for profit, and the final program displays
' the following words prominently:

' "This program is based on the PC Solutions CHECKER-KIT basic program."
' "(C) PC Solutions, Box 954,Bournemouth BH7 6YJ, England. You can purchase"
' "this source code, and many other stronger CHESS & CHECKERS programs from"
' "PC Solutions."

'   Remember the above only applies to .EXE files compiled with a compiler
' like the QUICKBASIC compiler - and not to the source .BAS file, or
' indeed any other PC Solutions program - these you must NOT copy!
'   This program is elementary - the emphasis is on simplicity, and I
' have left you plenty of room for improvement. The heart of the program
' is the subroutine COMPUTERMOVE, which recursively calls itself to
' search up and down the game tree. The search goes to a depth dictated
' by the FUEL variable - however the search does not terminate until
' jump sequences are completed. The most important search pruning
' method is the ALPHA-BETA cut - try setting the ALPHAMODE variable to
' zero to see how much slower the search is without it- but on a low level
' otherwise it will take forever! When ALPHAMODE is off, the computer
' operates a simple MINI-MAX search - ie. it looks at EVERY reply to every
' move upto the FUEL depth, then feeds back THESCORE variable (from terminal
' node) back down the tree, with White selecting the MAXIMUM score move
' and Black selecting the MINIMUM score at each ply-level. With ALPHAMODE
' on, a massive amount of the tree can be pruned with ALPHA-BETA pruning.
' This is a 'perfect' method - ie. it yields the same result as a MINI-MAX
' search. To understand ALPHA-BETA its best to look at an example.
' Consider a situation with white to play, where the computer has already
' looked at whites first move, which yields an even score of zero, kept
' in TBESTSCORE(ply). Now when the second white move is looked at, all
' blacks replys are generated and the first black reply is searched,
' and found to be a strong move, which wins material for Black, yielding
' a score of -100 at TBESTSCORE(ply+1). Now however good or bad subsequent
' black replys are, White will never play this move anyway because this
' black reply refutes the move. Since black has already found a move
' which refutes whites second move, no more black replys need to be looked
' at, and the search can cut back to Whites third move at (ply). So, in
' essence, by comparing TBESTSCORE(ply) with the value TBESTSCORE(ply-1),
' an ALPHA-BETA CUT can be made, often pruning massive amounts of the
' search tree off, yet yielding the same result as if the whole tree
' has been searched. The order moves are searched is critical - much can
' be achieved by placing potential good moves at the beginning of the search
' since this improves ALPHA-BETA pruning. The simplest method, the KILLER
' MOVE is implemented here - try KILLMODE = 0 to disable this. Many other
' methods could be used, such as ITERATIVE DEEPENING, to improve
' ordering - this method also allows you to adopt proper move-time
' regulation. At the moment, you will find great variation in the move time
' in different positions, since tree searching time is so unpredictable.
' To get round this, ITERATIVE DEEPENING, conducts searches at 1 ply, then
' 2 ply, 3 ply, etc - until time runs out and the best move so far is played.
' Information from shallower searches can be used to sort the moves in later
' searches into a good order. I cannot cover aspects of tree searching
' in depth here - I suggest you look for books and newsletters on this
' subject for more detail.
'   Another area that can be improved is the TERMINALNODE routine - at the
' moment this is very simple - the program plays for simple material
' balance, and for one or two parameters like the BRIDGE. The endgame is
' particularly weak - some more parameters for kings here would help a lot.
' Play will improve if you add the right parameters - ie the DOG-HOLE.
'   Of course, all these advanced methods and more are implemented in
' my SAGE/DYNAMO/CHESSMATE programs - however I don't want to give away
' all my trade secrets, and anyway the extra code complicationss would
' make the program code almost unreadable for a beginner.
'   Note that the notation used internally for the BOARD array is different
' from the traditional 1-32 square numbering - ie:
'  Ŀ
'   62 60 58 56   White side
'   55 53 51 49
'   46 44 42 40
'   39 37 35 33      Numbering used for BOARD () array.
'   30 28 26 24
'   23 21 19 17
'   14 12 10 8 
'   7  5  3  1    Black side
'  
'   By arranging in the way, move directions can be mode by simply adding
' or subtracting 7 or 9 to the current location - try it! You can also
' refer to locations by having XPOS and YPOS co-ordinates (0 to 7) then
' using:   CPOS = XPOS + YPOS * 8   to get a BOARD array location. In
' the above diagram, Traditional square 4 is numbered 7, 32 becomes 62, etc.
' See FNBRDTOSQ and FNSQTOBRD for conversion routines.
' Fans of the Italian/Spanish game could adapt this program for thier
' board layout, by simply changing the way FNPUTPIECE maps the squares
' to the screen.
' I hope you will forgive any minor spelling errors in the above - I started
' of doing this as a few simple REM's for the source code, and ended up
' writing a thesis! Well, as I said, I have deliberately left lots of room
' for improvement - I hope you can get lots of fun trying to improve
' CHECKER-KIT!

     DEFINT A-Z
		' Piece values on the BOARD array.
     Grey = -8: bman = 2: bking = 3: wman = 4: wking = 5

     maxply = 50                ' Absolute max depth of search
     maxlist = 1000             ' Max size of all move lists.
     maxbrd = 63                ' Size of board
     alphamode = 1              ' Alpha-beta on/off flag
     killmode = 1               ' Killer move flag
     randmode = 1               ' Randomise play a bit..
     moveptr = 0                ' Index to start of cur. move tab in lists
     fuel = 4                   ' Depth of search
     player = bman              ' Player to go, 2 =blk, 4= wht
     nomoves = 0                ' Number of moves available
     nonodes = 0                ' Number of term nodes
     RANDOMIZE TIMER
     valman = 100               ' Material weight..
     valking = 130
     ngame = 0                  ' No of moves played in mgame(), -1 disable
     bookflag = 1               ' Enable opening book
     numflag = 0                ' Number board positions..

     DIM board(100)     ' Board with pieces. 0 = blank, -8 = grey sq,
			' 2= bman, 3= bking, 4= wman, 5= wking
			' These values can be tested with simple AND cmds

     DIM listmov(1001)    ' Move list - loc of each movable piece
     DIM listdir(1001)    ' Dir list - direction of each piece
     DIM listeval(1001)   ' Temp eval - used for shortlook sort.

			' STACK VARIABLE - used to store current values
			' at a given ply during recursive search.

     DIM slevel(maxply)      ' Level of search.
     DIM svalof(maxply)      ' Val of moved piece, at any PLY
     DIM scaptured(maxply)   ' Val of captured piece, at any PLY
     DIM sjump(maxply)       ' Jump status, at any PLY
     DIM sscore(maxply)      ' Cumulative score, at any PLY
     DIM splayer(maxply)     ' Player to move, at any PLY
     DIM snomoves(maxply)    ' No of moves, at any PLY
     DIM smove(maxply)       ' Current Move no, at any PLY
     DIM sbestmov(maxply)    ' Best so far, at any PLY
     DIM sbestdir(maxply)    ' Best dir, at any PLY
     DIM sbestscore(maxply)  ' Best score, at any PLY

     DIM mgame(3000)     ' Moves in game so far (NGAME played)
			' Graphics arrays..
     DIM xbman(2000), xbking(2000), xwman(2000), xwking(2000), xgrey(2000)


     GOTO master                ' Run main prog.

DEF fnputat (xpos, ypos, cpiece)
  xpos = (7 - xpos) * charx + offx: ypos = (7 - ypos) * chary + offy
  IF cpiece = 0 THEN
    LINE (xpos, ypos)-(xpos + charx - 1, ypos + chary - 1), 0, BF
  ELSEIF cpiece = Grey THEN
    PUT (xpos, ypos), xgrey, PSET
  ELSEIF cpiece = bman THEN
    PUT (xpos, ypos), xbman, PSET
  ELSEIF cpiece = bking THEN
    PUT (xpos, ypos), xbking, PSET
  ELSEIF cpiece = wman THEN
    PUT (xpos, ypos), xwman, PSET
  ELSEIF cpiece = wking THEN
    PUT (xpos, ypos), xwking, PSET
  END IF
END DEF

				' Convert square # to board coord

DEF fnsqtobrd (x) = x * 2 - 1 + (((x - 1) AND 4) > 0)

				' Convert board coord to square #
DEF fnbrdtosq (x) = INT(x / 2) + 1

DEF fnmove$ (mloc, mdir) = MID$(STR$(fnbrdtosq(mloc)), 2) + "-" + MID$(STR$(fnbrdtosq(mloc + mdir)), 2)

DEF fnmovesq$ (sqf, sqt) = MID$(STR$(sqf), 2) + "-" + MID$(STR$(sqt), 2)

DEF fngametomove (vgame)       ' De-code move in GAME array
     fngametomove = fnbrdtosq(vgame \ 256) * 100 + fnbrdtosq(vgame AND 255)
END DEF

makemove:     ' Make legal move (xloc,xdir) for PLAYER
     moveptr = 0                 ' Init move pointer to start
     IF player = wman THEN
       GOSUB genwhitemoves     ' Gen list of legal white moves in LISTMOV/DIR
     ELSE
       GOSUB genblackmoves     ' Gen list of legal black moves in LISTMOV/DIR
     END IF
     badmove = 1            ' Set for illegal move.
     FOR move = 1 TO nomoves
       IF listmov(move) = xloc AND listdir(move) = xdir THEN badmove = 0
     NEXT move
     IF (badmove = 1) THEN RETURN    ' Illegal move..
     target = xloc + xdir
     IF ABS(xdir) > 10 THEN           ' Jump ?
       board(xloc + xdir / 2) = 0       ' Capture!
     END IF
     board(target) = board(xloc)
     board(xloc) = 0
     kingmade = 0
     IF target > 55 AND xdir > 0 THEN           ' Crown..
       kingmade = 1
       board(target) = board(target) OR 1
     END IF
     IF target < 8 AND xdir < 0 THEN
       board(target) = board(target) OR 1
       kingmade = 1
     END IF
     IF ngame > -1 THEN         ' Store move in game list..
       ngame = ngame + 1
       mgame(ngame) = xloc * 256 + target
     END IF
     RETURN


master:                   ' Main program loop - Input & execute commands
    CLS
    PRINT "CHECKER-KIT demo draughts program, (C) PC Solutions 1993."
    PRINT
    PRINT "This program is based on the PC Solutions CHECKER-KIT basic program."
    PRINT "(C) PC Solutions, Box 954,Bournemouth BH7 6YJ, England. You can purchase"
    PRINT "the source code of CHECKER-KIT, and many other stronger CHESS & CHECKERS"
    PRINT "programs from PC Solutions."
    PRINT
    PRINT "When using, CAPS LOCK OFF please!"
    PRINT "When playing, enter your commands or moves followed by RETURN, ie:"
    PRINT "   h <RETURN>     for HELP,"
    PRINT "   11-15 <RETURN> to play a move,"
    PRINT "   . <RETURN>     to display board numbering."
    PRINT "   Note - multiple jumps must be made IN STAGES - ie:-"
    PRINT "        11-18 <RETURN> 18-25 <RETURN> etc."
    PRINT
    PRINT "Do you want Text mode, or VGA graphics (t,v)"
    DO
      xgamode = INSTR(" tv", INKEY$) - 2
    LOOP WHILE (xgamode < 0)
    IF (xgamode) THEN GOSUB initgraphics
    CLS
    GOSUB newbrd
    human$ = ""
    WHILE (human$ <> "q")
      GOSUB printbrd
      COLOR 6
      LOCATE 20, 1
      PRINT "(h for help)"
      COLOR 7
      PRINT "Your Move?"
      PRINT "          "
      LOCATE 22, 1
      SOUND 200, 1
      IF human$ = "a" THEN      ' Auto play ?
	IF INKEY$ = CHR$(27) THEN human$ = ""
      END IF
      IF human$ <> "a" THEN LINE INPUT human$
      PRINT "               "
      PRINT "               ";
      LOCATE 23, 1
      IF VAL(human$) > 0 THEN       ' nn-nn move ?
	sqfrom = VAL(human$)
	sqto = VAL(MID$(human$, INSTR(human$, "-") + 1))
	xloc = fnsqtobrd(sqfrom)
	xdir = fnsqtobrd(sqto) - xloc
	GOSUB makemove     ' Do a move..
	IF badmove THEN
	  LOCATE 24, 1
	  PRINT "Bad Move!";
	ELSE                    ' Move OK, Double Jump?
	  GOSUB printbrd
	  jumpseq = 0
	  IF kingmade = 0 AND jump THEN    ' Test for multi jump sequence
	    gensq = xloc + xdir
	    GOSUB movesforgensq            ' Gen a move list for GENSQ
	    IF jump THEN
	      LOCATE 24, 1
	      PRINT "Jump Again!";
	      jumpseq = 1
	    END IF
	  END IF
	  IF jumpseq = 0 THEN           ' Not multi-jump, comp reply..
	    player = 6 - player
	    GOSUB docompmove
	    player = 6 - player
	  END IF
	END IF
      END IF
      IF human$ = "h" THEN
	CLS
	PRINT "#-#      Input your move - ie. 11-15 <RETURN>"
	PRINT "g        Make computer take YOUR turn- swapping sides."
	PRINT "l#       Set an play strength level (0-9), ie l5 <RETURN>"
	PRINT "x        Exchange side to play"
	PRINT "e        Enter pieces onto board"
	PRINT "c        Clear whole board"
	PRINT "n        New game - restart"
	PRINT "a        Auto-play, until ESC."
	PRINT "/a       Toggle alpha prune"
	PRINT "/k       Toggle killer move"
	PRINT "/b       Toggle opening book on/off"
	PRINT "/r       Toggle randomiser off/on."
	PRINT ".        Display board numbering."
	PRINT "q        Quit Checker-Mate."
	PRINT
	PRINT "         Hit SPACE to continue."
	WHILE INKEY$ <> " "
	WEND
	CLS
      END IF
      IF human$ = "/a" THEN      ' Alpha toggle
	alphamode = 1 - alphamode
	PRINT "Alpha = "; alphamode
      END IF
      IF human$ = "/k" THEN      ' killer toggle
	killmode = 1 - killmode
	PRINT "Kill = "; killmode
      END IF
      IF human$ = "/r" THEN
	randmode = 1 - randmode
	PRINT "Rand = "; randmode
      END IF
      IF human$ = "!" THEN
	CLS
	PRINT "Checker-Kit is Copyright and Written by A.Millett, PC Solutions."
	PRINT : PRINT "Hit SPACE key.."
	WHILE (INKEY$ <> " "): WEND
      END IF
      IF human$ = "g" OR human$ = "a" THEN      ' Move comp, Go..
	GOSUB docompmove
	player = 6 - player
      END IF
      IF human$ = "x" THEN      ' Change sides
	player = 6 - player
      END IF
      IF human$ = "/b" THEN     ' Book toggle
	bookflag = 1 - bookflag
	PRINT "Bookmode="; bookflag
      END IF
      IF LEFT$(human$, 1) = "l" THEN      ' Set level
	fuel = VAL(MID$(human$, 2))
      END IF
      IF human$ = "." THEN
	numflag = 1 - numflag
      END IF
      IF human$ = "$" THEN      ' Test movegen
	GOSUB testmoves
      END IF
      IF human$ = "c" THEN    ' Clear board
	GOSUB clrbrd
      END IF
      IF human$ = "n" THEN    ' Start new game
	GOSUB newbrd
      END IF
      IF human$ = "e" THEN      ' Enter pieces
	COLOR 7
	sq = 1
	WHILE sq > 0 AND sq < 33
	  CLS
	  GOSUB printbrd
	  LOCATE 15, 1
	  PRINT "0= empty, 2= bman,"
	  PRINT "3= bking, 4= wman,"
	  PRINT "5= wking"
	  PRINT "Enter square (1-32)"
	  INPUT sq
	  IF sq > 0 AND sq < 33 THEN
	    INPUT "Piece"; piece
	    board(fnsqtobrd(sq)) = piece
	  END IF
	WEND
	CLS : GOSUB printbrd
	ngame = -1              ' Ok, disable game list
      END IF
    WEND                ' All done, exit.
    END

gen1blksq:              ' Generate move list for 1 black square, mloc.

     IF mloc < 56 THEN          ' Move up and right ?
       dest = board(mloc + 7)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = 7
       ELSEIF (dest AND wman) > 0 AND mloc < 47 THEN
	 IF board(mloc + 14) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = 14
	 END IF
       END IF                   ' End for dir
     END IF

     IF mloc < 54 THEN          ' Move up and left ?
       dest = board(mloc + 9)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = 9
       ELSEIF (dest AND wman) > 0 AND mloc < 45 THEN
	 IF board(mloc + 18) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = 18
	 END IF
       END IF                   ' End for dir
     END IF

   IF board(mloc) = bking THEN      ' King move ?

     IF mloc > 7 THEN          ' Move Down left ?
       dest = board(mloc - 7)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = -7
       ELSEIF (dest AND wman) > 0 AND mloc > 16 THEN
	 IF board(mloc - 14) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = -14
	 END IF
       END IF                   ' End for dir
     END IF

     IF mloc > 9 THEN          ' Move Down right ?
       dest = board(mloc - 9)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = -9
       ELSEIF (dest AND wman) > 0 AND mloc > 18 THEN
	 IF board(mloc - 18) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = -18
	 END IF
       END IF                   ' End for dir
     END IF

   END IF               ' End for king move test
   RETURN


			' Generate move list for 1 white square, mloc.
gen1whtsq:

     IF mloc > 7 THEN          ' Move Down left ?
       dest = board(mloc - 7)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = -7
       ELSEIF (dest AND bman) > 0 AND mloc > 16 THEN
	 IF board(mloc - 14) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = -14
	 END IF
       END IF                   ' End for dir
     END IF

     IF mloc > 9 THEN          ' Move Down right ?
       dest = board(mloc - 9)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = -9
       ELSEIF (dest AND bman) > 0 AND mloc > 18 THEN
	 IF board(mloc - 18) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = -18
	 END IF
       END IF                   ' End for dir
     END IF

   IF board(mloc) = wking THEN      ' King move ?

     IF mloc < 56 THEN          ' Move up and right ?
       dest = board(mloc + 7)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = 7
       ELSEIF (dest AND bman) > 0 AND mloc < 47 THEN
	 IF board(mloc + 14) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = 14
	 END IF
       END IF                   ' End for dir
     END IF

     IF mloc < 54 THEN          ' Move up and left ?
       dest = board(mloc + 9)
       IF dest = 0 AND jump = 0 THEN
	 nomoves = nomoves + 1            ' Add to move list
	 listmov(moveptr + nomoves) = mloc
	 listdir(moveptr + nomoves) = 9
       ELSEIF (dest AND bman) > 0 AND mloc < 45 THEN
	 IF board(mloc + 18) = 0 THEN           ' Jump it ?
	   IF jump = 0 THEN             ' 1st jump, so clear table
	     jump = 1
	     nomoves = 0
	   END IF
	   nomoves = nomoves + 1          ' Add to move list
	   listmov(moveptr + nomoves) = mloc
	   listdir(moveptr + nomoves) = 18
	 END IF
       END IF                   ' End for dir
     END IF

   END IF               ' End for king move test
   RETURN


genblackmoves:          ' Gen list of legal black moves in LISTMOV/DIR
			'  offset by MOVEPTR. Return nomoves, jump.
     jump = 0
     nomoves = 0
     FOR mloc = 1 TO 62
       IF board(mloc) AND bman THEN GOSUB gen1blksq
     NEXT mloc
     RETURN

genwhitemoves:          ' Gen list of legal white moves in LISTMOV/DIR
			'  offset by MOVEPTR. Return nomoves, jump.
     jump = 0
     nomoves = 0
     FOR mloc = 1 TO 62
       IF board(mloc) AND wman THEN GOSUB gen1whtsq
     NEXT mloc
     RETURN

movesforgensq:          ' Gen a move list for GENSQ
	 jump = 0
	 nomoves = 0
	 mloc = gensq
	 IF player = wman THEN
	   GOSUB gen1whtsq
	 ELSE
	   GOSUB gen1blksq
	 END IF
	 RETURN


terminalnode:    ' Put evaluation function here - return THESCORE
     nonodes! = nonodes! + 1
     thescore = score
	' Remember to INCREASE this for WHITE gain, DECREASE for BLACK gain.
	' ie. Add 9 points if WHITE has the strong back-bridge.

			' BRIDGE squares
      IF board(1) = bman THEN IF board(5) = bman THEN thescore = thescore - 9
      IF board(62) = wman THEN IF board(58) = wman THEN thescore = thescore + 9
			' Keep Back row
      IF board(3) = bman THEN thescore = thescore - 3
      IF board(60) = wman THEN thescore = thescore + 3
			' Evacuate single corner
      IF board(7) = bman THEN thescore = thescore + 2
      IF board(56) = wman THEN thescore = thescore - 2
			' Kings in center
      'IF board(28) = bking THEN thescore = thescore - 4
      'IF board(35) = bking THEN thescore = thescore - 4
      'IF board(28) = wking THEN thescore = thescore + 4
      'IF board(35) = wking THEN thescore = thescore + 4
			' Add your own here...
     RETURN

'Ŀ
' 62 60 58 56   White side
' 55 53 51 49
' 46 44 42 40
' 39 37 35 33
' 30 28 26 24
' 23 21 19 17
' 14 12 10 8 
' 7  5  3  1    Black side
'

findkiller:             ' Search for killer, if found, put at top of list..
     killmove = sbestmov(ply)
     killdir = sbestdir(ply)
     FOR cmove = 1 TO nomoves            ' See if killer is there..
       IF listmov(moveptr + cmove) = killmove THEN
	 IF listdir(moveptr + cmove) = killdir THEN     ' Found it!
	   SWAP listmov(moveptr + 1), listmov(moveptr + cmove)
	   SWAP listdir(moveptr + 1), listdir(moveptr + cmove)
	   RETURN
	 END IF
       END IF
     NEXT cmove
     RETURN

'  The COMPUTERMOVE subroutine is the 'brains' of the program - it
' recursively examines each reply to each move upto FUEL level thus:-
'
'COMPUTERMOVE:
'  Generate List of legal moves (NOMOVES) for PLAYER (white/black)
'  IF LEVEL > FUEL THEN GOSUB TERMINALNODE and RETURN variable THESCORE
'  Save variables on stack
'  TBESTSCORE(PLY) = +/- 32767 - start with lowest possible score
'  FOR MOVE = 1 to NOMOVES - examine each move..
'    Execute move on BOARD
'    Save more variable on stack
'    PLY = PLY + 1
'    GOSUB COMPUTERMOVE
'    PLY = PLY - 1
'    Restore variables from stack
'    Undo move - put BOARD back as it was
'    if THESCORE is better than TBESTSCORE then Best move/score=current one.
'    if THESCORE is better than TBESTSCORE(PLY-1) then ALPHABETA cut.
'  NEXT MOVE
'  THESCORE = TBESTSCORE(PLY) and return this value.
'   

computermove:                    ' Recursive tree analysis.
     IF jumpseq = 0 THEN         ' Only Gen moves if not in multi jump sequence
       IF player = wman THEN      ' White ?
	 GOSUB genwhitemoves     ' Gen list of legal white moves in LISTMOV/DIR
	 IF nomoves = 0 THEN     ' No moves, all is lost!
	   thescore = -32700 + ply
	   nonodes! = nonodes! + 1
	   RETURN
	 END IF
       ELSE
	 GOSUB genblackmoves     ' Gen list of legal black moves in LISTMOV/DIR
	 IF nomoves = 0 THEN     ' No moves, all is lost!
	   thescore = 32700 - ply
	   nonodes! = nonodes! + 1
	   RETURN
	 END IF
       END IF
     END IF
     
     level = level + 1                    ' Now at deeper level..
     IF jump = 0 AND nomoves > 1 THEN     ' SEARCH TERMINATION TEST..
       IF level > fuel AND ply > 1 THEN   ' Out of fuel! So..
	 GOSUB terminalnode               ' Evaluate board position, THESCORE
	 RETURN                           ' Back to previous ply.
       END IF
     END IF
		' Store and Initialise stack for this ply.

     IF killmode AND nomoves > 1 THEN GOSUB findkiller      ' Killer move ?

     slevel(ply) = level        ' Level so far
     splayer(ply) = player      ' Save Player to move
     snomoves(ply) = nomoves    ' Save No of moves
     sjump(ply) = jump          ' Save Jump status
     sscore(ply) = score        ' Save Cumulative score
     sbestmov(ply) = 0          ' Save Best move so far
     sbestdir(ply) = 0          ' Save Best dir

		' SCAN MOVES..
     IF player = wman THEN
       sbestscore(ply) = -32767   ' Init Best score
     ELSE
       sbestscore(ply) = 32767    ' Init Best score
     END IF
     move = 1
     WHILE (move <= nomoves)    ' Ok, now look at all moves..
				' Exec current move on board ..
       IF ply = 1 THEN
	 IF randmode THEN
	   score = score + INT(RND(1) * 3)
	 END IF
       END IF
       smove(ply) = move        ' Save move #
       cloc = listmov(moveptr + move)
       cdir = listdir(moveptr + move)
       piece = board(cloc)
       svalof(ply) = piece      ' Save piece
       board(cloc) = 0
       target = cloc + cdir     ' Target square
       kingmade = 0             ' Flag indicates king crowned

       IF piece = wman AND target < 8 THEN         ' Crown white ?
	 score = score + 30                     ' Update material score
	 kingmade = 1
       END IF

       IF piece = bman AND target > 55 THEN        ' Crown black ?
	 score = score - 30                     ' Update material score
	 kingmade = 1
       END IF
		' Simplistic Advance king algorithm - improve on this!
       IF piece = bking AND cdir < 0 THEN
	 score = score - 1
       END IF
       IF piece = wking AND cdir > 0 THEN
	 score = score + 1
       END IF

       board(target) = piece OR kingmade

       captured = 0
       IF jump THEN                     ' Exec jump
	 capsq = cloc + cdir / 2
	 captured = board(capsq)

	 IF captured AND wman THEN         ' Update material score
	   IF captured = wman THEN         ' White taken ..
	     score = score - valman
	   ELSE
	     score = score - valking
	   END IF
	 ELSE
	   IF captured = bman THEN         ' Black taken ..
	     score = score + valman
	   ELSE
	     score = score + valking
	   END IF
	 END IF
	 board(capsq) = 0               ' Take piece off brd
       END IF

       scaptured(ply) = captured

       moveptr = moveptr + nomoves      ' Advance move pointer for new list
       jumpseq = 0
       IF kingmade = 0 AND jump THEN    ' Test for multi jump sequence
	 gensq = target
	 GOSUB movesforgensq            ' Gen a move list for GENSQ
	 IF jump THEN
	   jumpseq = 1                  ' Yes, multi jump seq (same PLAYER)
	 ELSE
	   player = 6 - player          ' Not multi, Change player
	 END IF
       ELSE
	 player = 6 - player            ' Not multi, Change player
       END IF

       ply = ply + 1
	 IF ply >= maxply THEN
	   PRINT "We've hit the ceiling!"
	   END
	 END IF

	 GOSUB computermove     ' Recursively examine deeper plys..
       ply = ply - 1


				   ' Restore vars for this ply.
       level = slevel(ply)         ' Level so far
       jump = sjump(ply)           ' Jump status
       score = sscore(ply)         ' Cumulative score
       player = splayer(ply)       ' Player to move
       nomoves = snomoves(ply)     ' No of moves

       moveptr = moveptr - nomoves ' Move Pointers back to old list

       move = smove(ply)
       captured = scaptured(ply)
       piece = svalof(ply)

				' Reverse move to restore BOARD ()

       cloc = listmov(moveptr + move)
       cdir = listdir(moveptr + move)
       board(cloc) = piece
       board(cloc + cdir) = 0
       IF captured THEN
	 board(cloc + cdir / 2) = captured
       END IF

       listeval(moveptr + move) = thescore      ' Save move val for short.

       IF player = wman THEN             ' Best white found ?
	 IF thescore > sbestscore(ply) THEN
	   IF alphamode AND ply > 1 AND player <> splayer(ply - 1) THEN
	     IF thescore >= sbestscore(ply - 1) THEN
	       RETURN                   ' Alpha cut, return THESCORE !
	     END IF
	   END IF
	   sbestmov(ply) = cloc
	   sbestdir(ply) = cdir
	   sbestscore(ply) = thescore
	 END IF
       ELSE                             ' Best black found ?
	 IF thescore < sbestscore(ply) THEN
	   IF alphamode AND ply > 1 AND player <> splayer(ply - 1) THEN
	     IF thescore <= sbestscore(ply - 1) THEN
	       RETURN                   ' Alpha cut, return THESCORE !
	     END IF
	   END IF
	   sbestmov(ply) = cloc
	   sbestdir(ply) = cdir
	   sbestscore(ply) = thescore
	 END IF
       END IF

       IF ply = 1 THEN
	 LOCATE 5, 1: PRINT move; ":"; fnmove$(cloc, cdir); "    ";
	 LOCATE 6, 1: PRINT "Eval:"; thescore; "   ";
	 LOCATE 7, 1: PRINT "Nodes:"; nonodes!; "  ";
       END IF
       move = move + 1          ' Next move..
     WEND                       ' Reloop until all moves looked at

     thescore = sbestscore(ply)         ' Pass best score to prev ply
     RETURN                             ' All done !

docompmove:
     COLOR 2
     nonodes! = 0               ' Number of term nodes
     timestart! = TIMER
     sbestmov(1) = 0            ' Clear killer move 1st ply
     doingshort = 0             ' Clr shortlook mode
     jumpseq = 0
     nonodes! = 0                ' Number of term nodes
     ply = 1
     level = 0
     score = 0
     moveptr = 0                 ' Init move pointer to start
     IF player = wman THEN
       GOSUB genwhitemoves     ' Gen list of legal white moves in LISTMOV/DIR
     ELSE
       GOSUB genblackmoves     ' Gen list of legal black moves in LISTMOV/DIR
     END IF
     IF nomoves = 0 THEN
       LOCATE 24, 1
       PRINT "I resign!";
       human$ = ""
       RETURN
     END IF
     IF bookflag AND ngame >= 0 THEN
       GOSUB getbookmove        ' Search book..
       IF BOOKMOVE THEN
	 bestmov = fnsqtobrd(BOOKMOVE \ 100)
	 bestdir = fnsqtobrd(BOOKMOVE MOD 100) - bestmov
	 bestscore = 0
	 GOTO docomp1           ' Found a book move, skip search..
       END IF
     END IF
compjumpagain:
     IF nomoves = 1 THEN
       bestmov = listmov(1)
       bestdir = listdir(1)
       bestscore = 0
     ELSE                       ' More than 1 move, so full search..
       GOSUB computermove
       bestmov = sbestmov(1)
       bestdir = sbestdir(1)
       bestscore = sbestscore(1)
     END IF

docomp1:
     extime! = TIMER - timestart!
     timetot! = timetot! + extime!
     LOCATE 5, 1: PRINT "Best:"; fnmove$(bestmov, bestdir); " ";
     LOCATE 6, 1: PRINT "Eval:"; bestscore; "   ";
     LOCATE 7, 1: PRINT "Nodes:"; nonodes!; "  ";
     LOCATE 8, 1: PRINT "Sec:"; INT(extime! * 10) / 10; "   "

     xloc = bestmov: xdir = bestdir: GOSUB makemove     ' Do a move..
     IF badmove THEN
       LOCATE 23, 1
       PRINT "Bad computer": PRINT "move ????"; : BEEP
     ELSE               ' Move OK, test for Multi-jump..
       GOSUB printbrd
       jumpseq = 0
       IF kingmade = 0 AND jump THEN    ' Test for multi jump sequence
	 gensq = xloc + xdir
	 GOSUB movesforgensq            ' Gen a move list for GENSQ
	 IF jump THEN
	   jumpseq = 1
	   GOTO compjumpagain
	 END IF
       END IF
     END IF
     RETURN

piececount:             ' Count how many pieces on board
     nmen = 0
     FOR csq = 1 TO 32
       cpiece = board(fnsqtobrd(csq))
       IF cpiece THEN nmen = nmen + 1
     NEXT csq
     RETURN

clrbrd:                         ' Set up clear board
     timetot! = 0
     ngame = -1                 ' Ok, disable game list
     FOR cloc = 0 TO 63
       board(cloc) = -8         ' Set all grey squares first
     NEXT cloc
     FOR csq = 1 TO 32          ' And clear active squares
       cloc = fnsqtobrd(csq)
       board(cloc) = 0
     NEXT csq
     RETURN

newbrd:                         ' Set up new brd
     player = bman                 ' Player to go, 2 =blk, 4= wht
     GOSUB clrbrd
     ngame = 0                  ' Ok, reset game list
     FOR csq = 1 TO 12
       cloc = fnsqtobrd(csq)
       board(cloc) = bman
     NEXT csq
     FOR csq = 21 TO 32
       cloc = fnsqtobrd(csq)
       board(cloc) = wman
     NEXT csq
     RETURN

printbrd:                               ' Print full board
   COLOR 14
   LOCATE 1, 1: PRINT "CHECKER-KIT"
   LOCATE 2, 1: PRINT "(C) PC Solutions."
   COLOR 10
   LOCATE 10, 1: PRINT "Level:"; fuel
   LOCATE 12, 1: PRINT MID$("Black White", player * 3 - 5, 5); " move.";
   IF xgamode = 0 THEN          ' Text mode..
     LOCATE 1, 40
     FOR bloc = 63 TO 0 STEP -1
       PRINT MID$("        ", 1 + (board(bloc) AND 15), 1);
       PRINT MID$("  " + CHR$(127) + CHR$(1) + CHR$(4) + CHR$(2) + "  ", 1 + (board(bloc) AND 15), 1);
       PRINT MID$("        ", 1 + (board(bloc) AND 15), 1);
       IF (bloc AND 7) = 0 THEN PRINT : LOCATE , 40
     NEXT bloc
     RETURN
   END IF
		' VGA Graphics mode.
   FOR cxpos = 0 TO 7
     FOR cypos = 0 TO 7
       temp = board(cxpos + cypos * 8)
       x = fnputat(cxpos, cypos, temp)  ' Onto VDU..
     NEXT cypos
   NEXT cxpos
   LINE (offx - 1, offy - 1)-(offx + charx * 8, offy + chary * 8), 7, B
   IF numflag THEN
     COLOR 7
     FOR csq = 1 TO 32
       cpos = fnsqtobrd(33 - csq)
       LOCATE 1 + INT(cpos / 8) * 3, 1 + (cpos AND 7) * 7 + offx / 8
       PRINT MID$(STR$(csq), 2);
     NEXT csq
   END IF
   RETURN

initgraphics:           ' Generate graphics for pieces..
    SCREEN 9
    charx = 56: chary = 41: offx = 160: offy = 2
    centx = charx / 2: centy = chary / 2
    asp! = .7
    LINE (1, 1)-(charx, chary), 8, BF
    GET (1, 1)-(charx, chary), xgrey
    LINE (1, 1)-(charx, chary), 0, BF
    CIRCLE (centx, centy), 20, 6, , , asp!
    PAINT (centx, centy), 6
    GET (1, 1)-(charx, chary), xbman
    CIRCLE (centx, centy), 14, 0, , , asp!
    CIRCLE (centx, centy), 7, 0, , , asp!
    GET (1, 1)-(charx, chary), xbking
    LINE (1, 1)-(charx, chary), 0, BF
    CIRCLE (centx, centy), 20, 7, , , asp!
    PAINT (centx, centy), 7
    GET (1, 1)-(charx, chary), xwman
    CIRCLE (centx, centy), 14, 0, , , asp!
    CIRCLE (centx, centy), 7, 0, , , asp!
    GET (1, 1)-(charx, chary), xwking
    SCREEN 9
    'PAINT (100, 100), 6
    RETURN

printmoves:                     ' Printout generated moves
     FOR cmove = 1 TO nomoves
       PRINT cmove; ":";
       PRINT fnmove$(listmov(moveptr + cmove), listdir(moveptr + cmove)),
     NEXT cmove
     PRINT
     RETURN

testmoves:                      ' Test move gen
    GOSUB genblackmoves
    PRINT "Black moves."
    GOSUB printmoves

    GOSUB genwhitemoves
    PRINT "White moves."
    GOSUB printmoves
    RETURN

getbookmove:    ' Compare MGAME(NGAME) with DATA statements,
		' and return BOOKMOVE, 0 if none found
    BOOKMOVE = 0
    RESTORE openbook
    READ dbook
    DO WHILE (dbook)    ' When zero, end of book.
      cgame = 1
      DO WHILE (cgame <= ngame AND fngametomove(mgame(cgame)) = dbook)
	cgame = cgame + 1
	READ dbook
      LOOP
      IF (cgame > ngame AND dbook) THEN  ' Matches book, move found..
	BOOKMOVE = dbook
	IF (RND(1) > .5) THEN
	  RETURN            ' 50% chance try another..
	END IF
      END IF
      DO WHILE (dbook)          ' Spool to next line
	READ dbook
      LOOP
      READ dbook
    LOOP
    RETURN

	' The computer simply searches the DATA until it finds
	' a matching line. Then there is a 50% chance of it taking
	' that move, or dropping through to the next line.
	' Note - because of this, the ordering is critical - put
	' weak moves at the bottom, so that the computer doesn't play it.

openbook:
    DATA 1115,0
    DATA 914,0
    DATA 1116,0
    DATA 1115,2218,0
    DATA 1115,2319,0
    DATA 1115,2217,0
    DATA 914,2218,509,2522,1116,1815,1019,2415,710,2724,1019,2415,1619,0
    DATA 914,2218,509,2419,1115,1811,815,1524,2819,711,0
    DATA 914,2217,1115,2522,811,1713,1116,2419,1524,2819,408,0
    DATA 1115,2319,811,2217,408,1713,1518,2420,1115,2824,811,2623,914,3126,609,0
    DATA 1115,2319,811,2217,913,1714,1017,2114,1518,1915,408,2419,0
    DATA 1115,2319,711,2623,307,2117,1116,2521,1620,3026,811,1916,1219,2316,913,2623,0
    DATA 1115,2217,811,2522,1116,2318,1519,2415,1019,0
    DATA 1115,2217,914,2522,811,1713,1116,2419,1524,2819,408,0
    DATA 1115,2217,1519,2415,1019,2316,1219,2522,811,3025,408,2218,1116,1714,812,0
    DATA 1115,2218,1522,2518,1216,2925,1014,2522,1620,2419,610,1916,812,2724,0
    DATA 1115,2218,1522,2518,811,2925,408,2420,1015,2522,1216,2117,812,1713,0
    DATA 1115,2318,811,2623,1014,3026,710,2419,1524,2819,1216,1912,1116,1815,0
    DATA 1115,2420,811,2824,913,2319,609,2723,914,3228,408,2218,0
    DATA 1116,2420,1619,2316,1219,2218,1014,1815,710,2522,1418,2925,0
    DATA 1116,2419,811,2218,1014,2622,1620,2217,710,2824,408,0
    DATA 1014,2218,1116,2622,710,2217,1619,2415,1019,1926,3023,811,0
    DATA 1015,2117,1116,1713,1620,2218,1522,2518,811,2925,0
    DATA 1216,2420,812,2824,914,2217,308,2622,1115,2011,716,2420,0
    DATA 913,2218,1015,2522,610,1814,1017,2114,1519,2415,1118,1825,3021,0
    DATA 0,0,0

